__author__ = "fccoelho"

from pathlib import Path
import unittest

import pandas as pd
from pysus.online_data.SIA import download
from pysus.online_data import parquets_to_dataframe

unittest.skip("too slow to run om travis")


class SIATestCase(unittest.TestCase):
    def test_download_after_2008(self):
        parquerts = download("to", 2015, 12)
        df = parquets_to_dataframe(parquerts)
        self.assertIn("PA_CODUNI", df.columns)
        self.assertIn("PA_GESTAO", df.columns)
        self.assertIsInstance(df, pd.DataFrame)
        self.assertIsInstance(df, pd.DataFrame)

    def test_download_before_2008(self):
        parquets = download("mg", 2005, 8)
        # self.assertWarns(UserWarning)
        df = parquets_to_dataframe(parquets)
        self.assertIn("PA_CODUNI", df.columns)
        self.assertIsInstance(df, pd.DataFrame)

    @unittest.expectedFailure
    def test_download_before_1994(self):
        file = download("RS", 1993, 12)
        self.assertTrue(Path(file).exists())

    def test_download_one(self):
        file = download("se", 2020, 10, group="PS")
        df = parquets_to_dataframe(file)
        self.assertIn("CNS_PAC", df.columns)
        self.assertIsInstance(df, pd.DataFrame)

    def test_download_many(self):
        files = []
        groups = ["aq", "AM", "atd"]
        for group in groups:
            files.extend([download("PI", 2018, 3, group=group)])
        to_df = parquets_to_dataframe    
        df1, df2, df3 = to_df(files[0]), to_df(files[1]), to_df(files[2])
        self.assertIsInstance(df1, pd.DataFrame)
        self.assertIsInstance(df2, pd.DataFrame)
        self.assertIsInstance(df2, pd.DataFrame)
        self.assertGreater(len(df1), 0)
        self.assertGreater(len(df2), 0)
        self.assertGreater(len(df3), 0)
        self.assertIn("AP_CODUNI", df1.columns)
        self.assertIn("AP_CODUNI", df2.columns)
        self.assertIn("AP_CODUNI", df3.columns)
        self.assertIn("AQ_CID10", df1.columns)
        self.assertIn("AM_PESO", df2.columns)
        self.assertIn("ATD_CARACT", df3.columns)

    def test_download_missing(self):
        dfs = parquets_to_dataframe(download("MS", 2006, 5))
        self.assertIsNotNone(dfs)


if __name__ == "__main__":
    unittest.main()



"""
Download SINASC data from DATASUS FTP server
Created on 01/11/17
by fccoelho
license: GPL V3 or Later
"""
from typing import Union

from pysus.online_data import CACHEPATH, FTP_Downloader, FTP_Inspect


def download(
    states: Union[str, list],
    years: Union[str, list, int],
    data_dir: str = CACHEPATH,
) -> list:
    """
    Downloads data directly from Datasus ftp server
    :param state: two-letter state identifier: MG == Minas Gerais,
                  can be a list
    :param year: 4 digit integer, can be a list
    :return: list of downloaded parquet paths
    """
    return FTP_Downloader("SINASC").download(
        UFs=states, years=years, local_dir=data_dir
    )


def get_available_years(state):
    return FTP_Inspect("SINASC").list_available_years(UF=state)



"""
Download of vacination data.

This module contains function to download from specific campains:

- COVID-19 in 2020-2021 Downloaded as described [here](http://opendatasus.saude.gov.br/dataset/b772ee55-07cd-44d8-958f-b12edd004e0b/resource/5916b3a4-81e7-4ad5-adb6-b884ff198dc1/download/manual_api_vacina_covid-19.pdf)
"""
import os
import json
import requests
import pandas as pd

from loguru import logger
from json import JSONDecodeError
from requests.auth import HTTPBasicAuth

from pysus.online_data import CACHEPATH


def download_covid(uf=None, only_header=False):
    """
    Download covid vaccination data for a give UF
    :param uf: 'RJ' | 'SP', etc.
    :param only_header: Used to see the header of the data before downloading.
    :return: dataframe iterator as returned by pandas `read_csv('Vaccine_temp_<uf>.csv.gz', chunksize=5000)`
    """
    user = "imunizacao_public"
    pwd = "qlto5t&7r_@+#Tlstigi"
    index = "desc-imunizacao"
    url = f"https://imunizacao-es.saude.gov.br/_search?scroll=1m"
    if uf is None:
        query = {"query": {"match_all": {}}, "size": 10000}
        UF = "BR"
    else:
        UF = uf.upper()
        query = {
            "query": {"match": {"paciente_endereco_uf": UF}},
            "size": 10000,
        }

    logger.info(f"Searching for COVID data of {UF}")
    tempfile = os.path.join(CACHEPATH, f"Vaccine_temp_{UF}.csv.gz")
    if os.path.exists(tempfile):
        print(
            "loading from cache. Returning an iterator of Dataframes in chunks of 5000."
        )
        return pd.read_csv(tempfile, chunksize=5000)

    auth = HTTPBasicAuth(user, pwd)
    data_gen = elasticsearch_fetch(url, auth, query)

    if only_header:
        df = pd.DataFrame(next(data_gen))
        logger.warning(
            f"Downloading data sample for visualization of {df.shape[0]} rows..."
        )
        return df

    h = 1
    for dt in data_gen:
        df = pd.DataFrame(dt)
        if h:
            df.to_csv(tempfile)
            h = 0
        else:
            df.to_csv(tempfile, mode="a", header=False)

    logger.info(f"{tempfile} stored at {CACHEPATH}.")
    df = pd.read_csv(tempfile, chunksize=5000)

    return df


def elasticsearch_fetch(uri, auth, json_body={}):
    headers = {
        "Content-Type": "application/json",
    }

    scroll_id = ""
    total = 0
    while True:
        if scroll_id:
            uri = "https://imunizacao-es.saude.gov.br/_search/scroll"
            json_body["scroll_id"] = scroll_id
            json_body["scroll"] = "1m"
            if "query" in json_body:
                del json_body[
                    "query"
                ]  # for the continuation of the download, query parameter is not allowed
                del json_body["size"]
        try:
            s = requests.Session()
            response = s.post(uri, auth=auth, headers=headers, json=json_body)
            text = response.text
            try:
                resp = json.loads(text)
            except JSONDecodeError:
                resp = text
        except Exception as error:
            print("\nelasticsearch_fetch() error:", error)
            raise error
        try:
            if resp["hits"]["hits"] == []:
                break
        except KeyError as e:
            logger.error(e)
            print(resp)
        total += len(resp["hits"]["hits"])
        print(f"Downloaded {total} records\r", end="")
        # print(resp)
        # print(uri)
        yield [h["_source"] for h in resp["hits"]["hits"]]
        if "_scroll_id" in resp:
            scroll_id = resp["_scroll_id"]


if __name__ == "__main__":
    print(download_covid("ba", only_header=True))



# type: ignore[attr-defined]
"""PySUS Python package"""

from . import preprocessing, utilities
from importlib import metadata as importlib_metadata


def get_version() -> str:
    try:
        return importlib_metadata.version(__name__)
    except importlib_metadata.PackageNotFoundError:  # pragma: no cover
        return "0.9.1"  # changed by semantic-release


version: str = get_version()
__version__: str = version



__author__ = "fccoelho"

import unittest
from unittest import skip

from pysus.online_data.SIM import (
    download,
    get_CID9_table,
    get_CID10_table,
    get_municipios,
    get_ocupations,
)
from pysus.online_data import parquets_to_dataframe as to_df

class TestDownload(unittest.TestCase):
    def test_download_CID10(self):
        df = to_df(download("ba", 2007))
        self.assertIn("IDADEMAE", df.columns)
        self.assertGreater(len(df), 0)

    def test_download_CID9(self):
        df = to_df(download("mg", 1987))
        self.assertIn("NECROPSIA", df.columns)
        self.assertGreater(len(df), 0)

    def test_get_cid10(self):
        df = get_CID10_table()
        self.assertIn("CID10", df.columns)
        self.assertGreater(len(df), 0)

    def test_get_cid9(self):
        df = get_CID9_table()
        self.assertIn("DESCRICAO", df.columns)
        self.assertGreater(len(df), 0)

    def test_get_mun(self):
        df = get_municipios()
        self.assertIn("MUNCOD", df.columns)
        self.assertGreater(len(df), 0)

    def test_get_ocup(self):
        df = get_ocupations()
        self.assertIn("CODIGO", df.columns)
        self.assertGreater(len(df), 0)


if __name__ == "__main__":
    unittest.main()



import unittest
from ftplib import FTP
import pandas as pd
from pysus.online_data.SIA import download
from pysus.online_data import parquets_to_dataframe as to_df

class SIATestCase(unittest.TestCase):
    @unittest.skip  # Takes a long time to complete
    def test_download_large_PA(self):
        res = to_df(download('SP', 2020, 12, group='PA'))
        if isinstance(res, pd.DataFrame):
            assert not res.empty
        else:
            pass


if __name__ == '__main__':
    unittest.main()



"""
This module contains a set of functions to decode
commonly encoded variables
Created on 19/07/16
by fccoelho
license: GPL V3 or Later
"""

__docformat__ = "restructuredtext en"
import re
from datetime import datetime, timedelta
from string import ascii_uppercase

import numpy as np
import pandas as pd

from pysus.online_data.SIM import (
    get_CID10_chapters_table,
    get_municipios,
)


@np.vectorize
def decodifica_idade_SINAN(idade, unidade: str = "Y"):
    """
    Em tabelas do SINAN frequentemente a idade é representada como um inteiro que precisa ser parseado
    para retornar a idade em uma unidade cronológica padrão.
    :param unidade: unidade da idade: 'Y': anos, 'M' meses, 'D': dias, 'H': horas
    :param idade: inteiro ou sequencia de inteiros codificados.
    :return:
    """
    fator = {"Y": 1.0, "M": 12.0, "D": 365.0, "H": 365 * 24.0}
    idade = int(idade)
    if idade >= 4000:  # idade em anos
        idade_anos = idade - 4000
    elif idade >= 3000 and idade < 4000:  # idade em meses
        idade_anos = (idade - 3000) / 12.0
    elif idade >= 2000 and idade < 3000:  # idade em dias
        idade_anos = (idade - 2000) / 365.0
    elif idade >= 1000 and idade < 2000:  # idade em horas
        idade_anos = (idade - 1000) / (365 * 24.0)
    else:
        idade_anos = np.nan
    idade_dec = idade_anos * fator[unidade]
    return idade_dec


def get_age_string(unidade):
    if unidade == "Y":
        return "ANOS"
    elif unidade == "M":
        return "MESES"
    elif unidade == "D":
        return "DIAS"
    elif unidade == "H":
        return "HORAS"
    elif unidade == "m":
        return "MINUTOS"
    else:
        return ""


@np.vectorize
def decodifica_idade_SIM(idade, unidade="D"):
    """
    Em tabelas do SIM a idade encontra-se codificada
    :param idade: valor original da tabela do SIM
    :param unidade: Unidade de saida desejada: 'Y': anos, 'M' meses, 'D': dias, 'H': horas, 'm': minutos. Valor default: 'D'
    :return:
    """
    fator = {"Y": 365.0, "M": 30.0, "D": 1.0, "H": 1 / 24.0, "m": 1 / 1440.0}
    try:
        if idade.startswith("0") and idade[1:] != "00":
            idade = timedelta(minutes=int(idade[1:]))
            idade = idade.seconds / 86400 + idade.days
        elif idade.startswith("1"):
            idade = timedelta(hours=int(idade[1:]))
            idade = idade.seconds / 86400 + idade.days
        elif idade.startswith("2"):
            idade = timedelta(days=int(idade[1:])).days
        elif idade.startswith("3"):
            idade = timedelta(days=int(idade[1:]) * 30).days
        elif idade.startswith("4"):
            idade = timedelta(days=int(idade[1:]) * 365).days
        elif idade.startswith("5"):
            idade = timedelta(days=int(idade[1:]) * 365).days + 100 * 365
        else:
            idade = np.nan
    except ValueError:
        idade = np.nan
    return idade / fator.get(unidade, 1)


@np.vectorize
def decodifica_data_SIM(data):
    try:
        new_data = datetime.strptime(data, "%d%m%Y")
    except ValueError:
        new_data = np.nan
    return new_data


@np.vectorize
def is_valid_geocode(geocodigo):
    """
    Returns True if the geocode is valid
    :param geocodigo:
    :return:
    """
    if len(str(geocodigo)) != 7:
        raise ValueError("Geocode must have 7 digtis")
    dig = int(str(geocodigo)[-1])
    if dig == calculate_digit(geocodigo):
        return True
    else:
        return False


def get_valid_geocodes():
    tab_mun = get_municipios()
    df = tab_mun[(tab_mun["SITUACAO"] != "IGNOR")]
    return df["MUNCODDV"].append(df["MUNCOD"]).astype("int64").values


def calculate_digit(geocode):
    """
    Calcula o digito verificador do geocódigo de município com 6 dígitos
    :param geocode: geocódigo com 6 dígitos
    :return: dígito verificador
    """
    peso = [1, 2, 1, 2, 1, 2, 0]
    soma = 0
    geocode = str(geocode)
    for i in range(6):
        valor = int(geocode[i]) * peso[i]
        soma += sum([int(d) for d in str(valor)]) if valor > 9 else valor
    dv = 0 if soma % 10 == 0 else (10 - (soma % 10))
    return dv


@np.vectorize
def add_dv(geocodigo):
    if len(str(geocodigo)) == 7:
        return geocodigo
    else:
        return int(str(geocodigo) + str(calculate_digit(geocodigo)))


def columns_as_category(series, nan_string=None):
    series = series.astype("category")
    series = series.cat.add_categories


def translate_variables_SIM(
        dataframe: pd.DataFrame,
        age_unit: str = "Y",
        age_classes=None,
        classify_args: dict = {},
        classify_cid10_chapters=False,
        geocode_dv=True,
        nan_marker=None,
        category_columns=True,
):
    variables_names = dataframe.columns.tolist()
    df = dataframe

    valid_mun = get_valid_geocodes()

    # IDADE
    if "IDADE" in variables_names:
        column_name = f"IDADE_{get_age_string(age_unit)}"
        df[column_name] = decodifica_idade_SIM(df["IDADE"], age_unit)
        if age_classes:
            df[column_name] = classify_age(df[column_name], **classify_args)
            df[column_name] = df[column_name].astype("category")
            df[column_name] = df[column_name].cat.add_categories(["NA"])
            df[column_name] = df[column_name].fillna("NA")

    # SEXO
    if "SEXO" in variables_names:
        df['SEXO'] = df.SEXO.replace(
            {0: None, 9: None, 1: "Masculino", 2: "Feminino"}
        )
        df["SEXO"] = df["SEXO"].astype("category")
        df["SEXO"] = df["SEXO"].cat.add_categories(["NA"])
        df["SEXO"] = df["SEXO"].fillna("NA")

    # MUNRES
    if "MUNIRES" in variables_names:
        df = df.rename(columns={"MUNIRES": "CODMUNRES"})
        variables_names.append("CODMUNRES")

    # CODMUNRES
    if "CODMUNRES" in variables_names:
        if geocode_dv:
            df["CODMUNRES"] = df["CODMUNRES"].apply(add_dv)
        df["CODMUNRES"] = df["CODMUNRES"].astype("int64")
        df.loc[~df["CODMUNRES"].isin(valid_mun), "CODMUNRES"] = pd.NA
        df["CODMUNRES"] = df["CODMUNRES"].astype("category")
        df["CODMUNRES"] = df["CODMUNRES"].cat.add_categories(["NA"])
        df["CODMUNRES"] = df["CODMUNRES"].fillna("NA")

    # RACACOR
    if "RACACOR" in variables_names:
        df["RACACOR"] = df["RACACOR"].str.strip().replace(
            {
                "0": None,
                "1": "Branca",
                "2": "Preta",
                "3": "Amarela",
                "4": "Parda",
                "5": "Indígena",
                "6": None,
                "7": None,
                "8": None,
                "9": None,
                "": None,
            },
        )
        df["RACACOR"] = df["RACACOR"].astype("category")
        df["RACACOR"] = df["RACACOR"].cat.add_categories(["NA"])
        df["RACACOR"] = df["RACACOR"].fillna("NA")

    # CAUSABAS IN CID10 CHAPTER
    if classify_cid10_chapters:
        code_index = get_CID10_code_index(get_CID10_chapters_table())
        df["CID10_CHAPTER"] = df["CAUSABAS"].str.slice(0, 3).map(code_index)
        df["CID10_CHAPTER"] = df["CID10_CHAPTER"].astype("category")

    return df


def classify_age(
    serie, start=0, end=90, freq=None, open_end=True, closed="left", interval=None
):
    """
    Classifica idade segundo parâmetros ou IntervalIndex
    :param serie: Serie pandas contendo idades
    :param start: início do primeiro grupo
    :param end: fim do último grupo
    :param freq: tamanho dos grupos. Por padrão considera cada valor um grupo.
    :param open_end: cria uma classe no final da lista de intervalos que contém todos acima daquele último valor. Default True
    :param closed: onde os intervalos devem ser fechados. Possíveis valores: {'left', 'right', 'both', 'neither'}. Default 'left'
    :param interval: IntervalIndex do pandas. Caso seja passado todos os outros parâmetros de intervalo são desconsiderados. Defaul None
    :return:
    """
    if interval:
        iv = interval
    else:
        iv = pd.interval_range(start=start, end=end, freq=freq, closed=closed)
    iv_array = iv.to_tuples().tolist()

    # Adiciona classe aberta no final da lista de intervalos.
    # Útil para criar agrupamentos como 0,1,2,...,89,90+
    if open_end:
        iv_array.append((iv_array[-1][1], +np.inf))
    intervals = pd.IntervalIndex.from_tuples(iv_array, closed=closed)
    return pd.cut(serie, intervals)


def get_CID10_code_index(datasus_chapters):
    code_index = {}
    for ch_array_index, chapter in datasus_chapters.iterrows():
        # Ex.: ['A00','B99']
        chapter_range = chapter["CAUSAS"].split("-")
        start_letter = chapter_range[0][0]
        end_letter = chapter_range[1][0]

        if start_letter == end_letter:
            number_range_start = int(chapter_range[0][1:3])
            number_range_finish = int(chapter_range[1][1:3])
            for code in range(number_range_start, number_range_finish + 1):
                code_index[f"{start_letter}{str(code).zfill(2)}"] = ch_array_index + 1
        else:
            string_range_start = chapter_range[0][0]
            string_range_end = chapter_range[1][0]
            full_string_range = re.compile(
                f"{string_range_start}.*{string_range_end}"
            ).search(ascii_uppercase)[0]

            for let_array_index, letter in enumerate(full_string_range):
                # First array letter
                if let_array_index == 0:
                    number_range_start = int(chapter_range[0][1:3])
                    number_range_end = 99
                elif let_array_index == len(full_string_range) - 1:  # Last array letter
                    number_range_start = 0
                    number_range_end = int(chapter_range[1][1:3])
                else:  # Middle letters
                    number_range_start = 0
                    number_range_end = 99
                for code_number in range(number_range_start, number_range_end + 1):
                    code_index[f"{letter}{str(code_number).zfill(2)}"] = (
                        ch_array_index + 1
                    )

    return code_index



import datetime
import os
from pathlib import Path
import shutil
import unittest
from glob import glob
from pathlib import Path

import numpy as np
import pandas as pd

from pysus.online_data.SINAN import (
    download,
    list_diseases,
    metadata_df
)
from pysus.online_data import FTP_SINAN, parquets_to_dataframe
from pysus.preprocessing.sinan import read_sinan_dbf

PATH_ROOT = Path(__file__).resolve().parent


class TestSINANClass(unittest.TestCase):
    data_path = '/tmp/pysus'
    d1 = 'Raiva Humana'
    r1 = [
        'RAIVBR07.parquet',
        'RAIVBR08.parquet',
        'RAIVBR09.parquet',
        'RAIVBR10.parquet',
        'RAIVBR11.parquet',
        'RAIVBR12.parquet',
        'RAIVBR13.parquet',
        'RAIVBR14.parquet',
        'RAIVBR15.parquet',
        'RAIVBR16.parquet',
        'RAIVBR17.parquet',
        'RAIVBR18.parquet',
        'RAIVBR19.parquet',
    ]

    def test_list_all_diseases(self):
        all_diseases = list(FTP_SINAN.diseases.keys())
        self.assertIn('Dengue', all_diseases)
        self.assertIn('Zika', all_diseases)
        self.assertIn('Chikungunya', all_diseases)

    def test_download(self):
        files = download(self.d1, [7,8,9], data_path=self.data_path)
        self.assertEqual(len(files), 3)

    def test_read_dataframe(self):
        df = parquets_to_dataframe(Path(self.data_path)/self.r1[0])
        self.assertIsInstance(df, pd.DataFrame)
        self.assertEqual(df.shape, (110, 94))
    
    def test_metadata_dataframe(self):
        df = metadata_df('Raiva Humana')
        self.assertIsInstance(df, pd.DataFrame)
        self.assertEqual(df.shape, (68, 7))


class TestSINANDownload(unittest.TestCase):
    def test_download(self):
        df = parquets_to_dataframe(download(years=2007, disease='Botulismo'))
        self.assertIsInstance(df, pd.DataFrame)

    def test_filename_only(self):
        fname = download(years=2015, disease='Botulismo')
        self.assertIsInstance(fname, str)
        self.assertTrue(os.path.exists(fname))
        shutil.rmtree(fname, ignore_errors=True)

    def test_fetch_viol_dom(self):
        df = parquets_to_dataframe(download(years=2011, disease='Hantavirose'))
        self.assertIsInstance(df, pd.DataFrame)

    def test_fetch_cancer_prelim(self):
        df = parquets_to_dataframe(download(years=2022, disease='Cancer'))
        self.assertIsInstance(df, pd.DataFrame)

    def test_fetch_sifilis(self):
        self.assertRaises(
            Exception, download(years=2021, disease='Sífilis Adquirida')
        )

    def test_fetch_sifilis_gestante(self):
        df = parquets_to_dataframe(download(years=2021, disease='Sífilis em Gestante'))
        self.assertIsInstance(df, pd.DataFrame)

    def test_lista_agravos(self):
        lista = list_diseases()
        self.assertIsInstance(lista, list)
        self.assertGreater(len(lista), 0)

    def test_chunked_df_size(self):
        df1 = parquets_to_dataframe(download(years=2018, disease='Chikungunya'))
        s1 = len(df1)
        del df1
        fn = download(years=2018, disease='Chikungunya')
        for i, f in enumerate(glob(f'{fn}/*.parquet')):
            if i == 0:
                df2 = pd.read_parquet(f)
            else:
                df2 = pd.concat([df2, pd.read_parquet(f)], ignore_index=True)
        self.assertEqual(s1, df2.shape[0])
        shutil.rmtree(fn, ignore_errors=True)


class TestSinanDBF(unittest.TestCase):
    dbf_name = PATH_ROOT / 'EPR-2016-06-01-2016.dbf'

    def test_read_dbf(self):
        df = read_sinan_dbf(self.dbf_name, encoding='latin-1')
        self.assertTrue(self.dbf_name.exists())
        self.assertIsInstance(df, pd.DataFrame)
        for cname in df.columns:
            if cname.startswith('DT_'):
                self.assertIsInstance(df[cname][0], datetime.date)
            elif cname.startswith('SEM'):
                self.assertLessEqual(df[cname][0], 52)
                self.assertIsInstance(df[cname][0], (int, np.int64))
            elif cname.startswith(('NU', 'ID')):
                if cname == 'ID_AGRAVO':
                    continue
                self.assertIsInstance(
                    df[cname][0],
                    (int, float, np.int64),
                    msg='Failed on column {}, type:{}'.format(
                        cname, type(df[cname][0])
                    ),
                )

    def test_type_convertion(self):
        df = read_sinan_dbf(self.dbf_name, encoding='latin-1')
        self.assertTrue(self.dbf_name.exists())
        assert not all(df.dtypes == 'object')


if __name__ == '__main__':
    unittest.main()



import pandas as pd
from pathlib import Path
from typing import Union
from pysus.online_data import FTP_Downloader, FTP_Inspect, CACHEPATH, FTP_SINAN


def list_diseases() -> list:
    """List available diseases on SINAN"""
    return list(FTP_SINAN.diseases.keys())


def get_available_years(disease: str) -> list:
    """
    Fetch available years for data related to specific disease
    :param disease: Disease name. See `SINAN.list_diseases` for valid names
    :return: A list of DBC files from a specific disease found in the FTP Server.
    """
    return FTP_Inspect("SINAN").list_available_years(SINAN_disease=disease)


def download(
    disease, years: Union[str, list, int], data_path: str = CACHEPATH
) -> list:
    """
    Downloads SINAN data directly from Datasus ftp server.
    :param disease: Disease according to `agravos`.
    :param years: 4 digit integer, can be a list of years.
    :param data_path: The directory where the chunks will be downloaded to.
    :return: list of downloaded parquet directories.
    """
    return FTP_Downloader("SINAN").download(
        SINAN_disease=disease, years=years, local_dir=data_path
    )


def metadata_df(disease: str) -> pd.DataFrame:
    code = FTP_SINAN(disease).code
    metadata_file = (
        Path(__file__).parent.parent / "metadata" / "SINAN" / f"{code}.tar.gz"
    )
    if metadata_file.exists():
        df = pd.read_csv(
            metadata_file,
            compression="gzip",
            header=0,
            sep=",",
            quotechar='"',
            error_bad_lines=False,
        )

        return df.iloc[:, 1:]
    else:
        print(f"No metadata available for {disease}")
        return



import unittest

import pandas as pd

from pysus.online_data.CNES import download
from pysus.online_data import parquets_to_dataframe as to_df


class CNESTestCase(unittest.TestCase):
    @unittest.skip('Also fails in previous versions: unpack requires a buffer of 32 bytes')
    def test_fetch_estabelecimentos(self):
        df = to_df(download(group="ST", states="SP", years=2021, months=8))
        self.assertIsInstance(df, pd.DataFrame)
        # self.assertEqual(True, False)  # add assertion here

    def test_fetch_equipamentos(self):
        df = to_df(download(group="EQ", states="RO", years=2021, months=9))
        self.assertIsInstance(df, pd.DataFrame)


# if __name__ == "__main__":
#     unittest.main()



"""
Download data from the national immunization program
"""
from typing import Union

from pysus.online_data import CACHEPATH, FTP_Downloader, FTP_Inspect


def download(
    states: Union[str, list],
    years: Union[str, list, int],
    data_dir: str = CACHEPATH,
) -> list:
    """
    Download imunization records for a given States and years.
    :param state: uf two letter code, can be a list
    :param year: year in 4 digits, can be a list
    :param data_dir: directory where data will be downloaded
    :return: list of downloaded parquet paths
    """
    return FTP_Downloader("PNI").download(
        PNI_group="CPNI", UFs=states, years=years, local_dir=data_dir
    )


def get_available_years(state):
    """
    Fetch available years (dbf names) for the `state`.
    :param state: uf code
    :return: list of strings (filenames)
    """
    return FTP_Inspect("PNI").list_available_years(UF=state, PNI_group="CPNI")


def available_docs():
    return FTP_Inspect("PNI").list_all(PNI_group="CPNI")



"""
Created on 21/09/18
by fccoelho
license: GPL V3 or Later
"""
import logging
import os
import re
import shutil
from datetime import datetime
from ftplib import FTP, error_perm
from itertools import product
from pathlib import Path, PosixPath
from typing import Union

import pandas as pd
import pyarrow as pa
import pyarrow.parquet as pq
from dbfread import DBF
from pysus.utilities.readdbc import dbc2dbf

CACHEPATH = os.getenv(
    "PYSUS_CACHEPATH", os.path.join(str(Path.home()), "pysus")
)

# create pysus cache directory
if not os.path.exists(CACHEPATH):
    os.mkdir(CACHEPATH)


DB_PATHS = {
    "SINAN": [
        "/dissemin/publicos/SINAN/DADOS/FINAIS",
        "/dissemin/publicos/SINAN/DADOS/PRELIM",
    ],
    "SIM": [
        "/dissemin/publicos/SIM/CID10/DORES",
        "/dissemin/publicos/SIM/CID9/DORES",
    ],
    "SINASC": [
        "/dissemin/publicos/SINASC/NOV/DNRES",
        "/dissemin/publicos/SINASC/ANT/DNRES",
    ],
    "SIH": [
        "/dissemin/publicos/SIHSUS/199201_200712/Dados",
        "/dissemin/publicos/SIHSUS/200801_/Dados",
    ],
    "SIA": [
        "/dissemin/publicos/SIASUS/199407_200712/Dados",
        "/dissemin/publicos/SIASUS/200801_/Dados",
    ],
    "PNI": ["/dissemin/publicos/PNI/DADOS"],
    "CNES": ["dissemin/publicos/CNES/200508_/Dados"],
    "CIHA": ["/dissemin/publicos/CIHA/201101_/Dados"],
}

def FTP_datasus():
    ftp = FTP("ftp.datasus.gov.br")
    ftp.login()
    return ftp


def cache_contents():
    """
    List the files currently cached in ~/pysus
    :return:
    """
    cached_data = os.listdir(CACHEPATH)
    return [os.path.join(CACHEPATH, f) for f in cached_data]


def parquets_to_dataframe(
    parquet_dir: str(PosixPath), clean_after_read=False
) -> pd.DataFrame:
    """
    Receives a parquet directory path and returns it as a
    dataframe, trying to clean white spaces and convert to
    the correct data types. Can read only one parquet dir
    at time.
    """

    parquets = list(map(str, Path(parquet_dir).glob("*.parquet")))
    
    try:
        chunks_list = [
            pd.read_parquet(str(f), engine="fastparquet") for f in parquets
        ]
        df = pd.concat(chunks_list, ignore_index=True)

        return _parse_dftypes(df)

    except Exception as e:
        logging.error(e)

    finally:
        if clean_after_read:
            shutil.rmtree(parquet_dir)
            logging.info(f"{parquet_dir} removed")


def _parse_dftypes(df: pd.DataFrame) -> pd.DataFrame:
    """
    Parse DataFrame values, cleaning blank spaces if needed
    and converting dtypes into correct types.
    """

    def map_column_func(column_names: list[str], func):
        # Maps a function to each value in each column
        columns = [c for c in df.columns if c in column_names]
        df[columns] = df[columns].applymap(func)

    def str_to_int(string: str) -> Union[int, float]:
        # If removing spaces, all characters are int,
        # return int(value). @warning it removes in between
        # spaces as wel 
        if str(string).replace(" ", "").isnumeric():
            return int(string.replace(" ", ""))

    def str_to_date(string: str) -> datetime.date:
        if isinstance(string, str):
            try:
                return datetime.strptime(string, "%Y%m%d").date()
            except Exception:
                # Ignore errors, bad value
                pass

    map_column_func(["DT_NOTIFIC", "DT_SIN_PRI"], str_to_date)
    map_column_func(["CODMUNRES", "SEXO"], str_to_int)

    df = df.applymap(
        lambda x: "" if str(x).isspace() else x
    )  # Remove all space values

    df = df.convert_dtypes()
    return df


class FTP_Inspect:
    """
    Databases: "SINAN", "SIM", "SINASC", "SIH", "SIA", "PNI", "CNES", "CIHA"
    FTP_Inspect will focus mainly on enter in DataSUS ftp server
    and list the DBCs or DBFs paths for a database according to
    DB_PATH dict. Receives a Database as parameter.

    Methods
    last_update_df: Returns a DataFrame with information of the last
                    update from a database (Legacy) .

    list_available_years: Lists years found for a Database. Some DBs
                    contain groups that are needed to be passed in.

    list_all: Will list all DBC or DBF urls found on the FTP server
                    for the Database. Groups may be also required.
    """

    database: str
    _ds_paths: list
    ftp_server: FTP = FTP("ftp.datasus.gov.br")
    available_dbs: list = list(DB_PATHS.keys())

    def __init__(self, database: str) -> None:
        self.database = self.__checkdatabase__(database)
        self._ds_paths = DB_PATHS[database]

    def __checkdatabase__(self, database):
        if database not in self.available_dbs:
            raise ValueError(
                f"{database} not found"
                f" available databases: {self.available_dbs}"
            )
        return database

    def last_update_df(self) -> pd.DataFrame:  # Legacy
        """
        Return the date of last update from the database specified.

        Parameters
        ----------
        database: Database to check
        """
        if self.database not in DB_PATHS:
            print(
                f"Database {self.database} not supported try one of these"
                "{list(DB_PATHS.keys())}"
            )
            return pd.DataFrame()

        ftp = FTP_datasus()
        response = {
            "folder": [],
            "date": [],
            "file_size": [],
            "file_name": [],
        }

        def parse(line):
            data = line.strip().split()
            response["folder"].append(pth)
            response["date"].append(
                pd.to_datetime(" ".join([data[0], data[1]]))
            )
            response["file_size"].append(
                0 if data[2] == "<DIR>" else int(data[2])
            )
            response["file_name"].append(data[3])

        for pth in DB_PATHS[self.database]:
            ftp.cwd(pth)
            flist = ftp.retrlines("LIST", parse)
        
        ftp.close()
        return pd.DataFrame(response)

    def list_available_years(
        self,
        UF: str = None,
        SINAN_disease: str = None,
        CNES_group: str = None,
        SIA_group: str = "PA",
        PNI_group: str = "CPNI",
        SIH_group: str = "RD",
    ):
        """
        Uses `list_all` and filters according to UF, disease (SINAN),
        or Database group if group is required.
        """
        available_years = set()
        get_filename = (
            lambda x: str(x)
            .split("/")[-1]
            .upper()
            .split(".DBC")[0]
            .split(".DBF")[0]
        )  # Trim url paths

        def list_years(
            len_group: int, fslice: slice = slice(-2, None), **kwargs
        ):
            return [
                available_years.add(get_filename(path)[fslice])
                for path in self.list_all(**kwargs)
                if UF in get_filename(path)[len_group:]
            ]

        if UF is not None and len(UF) > 2:
            raise ValueError("Use UF abbreviation. Eg: RJ")

        # SINAN
        if self.database == "SINAN":
            if not SINAN_disease:
                raise ValueError("No disease assigned to SINAN_disease")
            dis = FTP_SINAN(SINAN_disease)
            available_years = dis.get_years(stage="all")
        # SINASC
        elif self.database == "SINASC":
            list_years(2)
        # SIH
        elif self.database == "SIH":
            list_years(len(SIH_group), slice(-4, -2), SIH_group=SIH_group)

        # SIA
        elif self.database == "SIA":
            list_years(len(SIA_group), slice(-4, -2), SIA_group=SIA_group)
        # CNES
        elif self.database == "CNES":
            list_years(len(CNES_group), slice(-4, -2), CNES_group=CNES_group)
        # PNI
        elif self.database == "PNI":
            list_years(len(PNI_group), PNI_group=PNI_group)
        # CIHA
        elif self.database == "CIHA":
            list_years(4)
        # SIM
        elif self.database == "SIM":
            dbcs = self.list_all()
            available_years = set()
            for path in dbcs:
                if "/CID9/" in path:
                    available_years.add(get_filename(path)[-2:]) if str(path)[
                        -8:-6
                    ] == UF else None
                elif "/CID10/" in path:
                    available_years.add(get_filename(path)[-2:]) if str(path)[
                        -10:-8
                    ] == UF else None

        # Normalize years to {year:04d} and return sorted
        cur_year = str(datetime.now().year)[-2:]
        bef_2000 = lambda yrs: [
            "19" + y for y in yrs if y > cur_year and y <= "99"
        ]
        aft_2000 = lambda yrs: [
            "20" + y for y in yrs if y <= cur_year and y >= "00"
        ]
        return sorted(bef_2000(available_years)) + sorted(
            aft_2000(available_years)
        )

    def list_all(
        self,
        SINAN_disease: str = None,
        CNES_group: str = None,
        SIA_group: str = "PA",
        PNI_group: str = "CPNI",
        SIH_group: str = "RD",
    ) -> list:
        """
        Enters FTP server and lists all DBCs or DBFs files found for a
        Database group. Some Database require groups and SINAN DB requires
        a disease, more details can be found in their modules.
        This method will be later used to download these files into parquets
        chunks, to preserve memory, that are read using pandas and pyarrow.
        """
        available_dbs = list()
        ftp = FTP("ftp.datasus.gov.br")
        ftp.login()
        for path in self._ds_paths:
            try:
                # CNES
                if self.database == "CNES":
                    if not CNES_group:
                        raise ValueError(f"No group assigned to CNES_group")
                    available_dbs.extend(
                        ftp.nlst(f"{path}/{CNES_group}/*.DBC")
                    )
                # SIA
                elif self.database == "SIA":
                    if not SIA_group:
                        raise ValueError(f"No group assigned to SIA_group")
                    available_dbs.extend(ftp.nlst(f"{path}/{SIA_group}*.DBC"))
                # SIH
                elif self.database == "SIH":
                    if not SIH_group:
                        raise ValueError(f"No group assigned to SIH_group")
                    available_dbs.extend(ftp.nlst(f"{path}/{SIH_group}*.DBC"))
                # PNI
                elif self.database == "PNI":
                    if not PNI_group:
                        raise ValueError(f"No group assigned to PNI_group")
                    available_dbs.extend(ftp.nlst(f"{path}/{PNI_group}*.DBF"))
                # SINAN
                elif self.database == "SINAN":
                    if not SINAN_disease:
                        raise ValueError(
                            f"No disease assigned to SINAN_disease"
                        )
                    disease = FTP_SINAN(SINAN_disease)
                    available_dbs = disease.get_ftp_paths(
                        disease.get_years(stage="all")
                    )
                # SIM, SINASC
                else:
                    available_dbs.extend(
                        ftp.nlst(f"{path}/*.DBC")  # case insensitive
                    )
            except Exception as e:
                raise e
        ftp.close()
        return available_dbs


class FTP_Downloader:
    """
    Databases: "SINAN", "SIM", "SINASC", "SIH", "SIA", "PNI", "CNES", "CIHA"
    FTP_Downloader will be responsible for fetching DBF and DBC files
    into parquet chunks, according to a DataSUS Database (DB_PATHS).
    The main function is `download`, each Database has its specific
    url pattern, some may require a group or disease (SINAN), some may
    not require a month, year nor UF. Independent the requirements, the
    group is the only that won't accept to passed in as list. A local
    directory can be set, default dir is CACHEPATH.

    Methods
    download: Filters the files from the FTP Database according to its
              specs (UFs, Years, Months, Disease &/or Group and local dir).
              The parametes has to be set using their names in the function
              with the equals sign. It will fetch a DBC or DBF file and parse
              them into parquet chunks that will be read using pandas.
              Example:
                ciha = FTP_Downloader('CIHA')
                ufs = ['RJ', 'AC']
                years = [2022, 2023]
                months = [1, 2, 3]
                ciha.download(UFs=ufs, years=years, months=months)
    """

    _ftp_db: FTP_Inspect
    dbc_paths: list = None
    cache_dir: str = CACHEPATH

    def __init__(self, database: str) -> None:
        self._ftp_db = FTP_Inspect(database)

    def download(
        self,
        UFs: Union[str, list] = None,
        years: Union[str, int, list] = None,
        months: Union[str, int, list] = None,
        SINAN_disease: str = None,
        CNES_group: str = None,
        SIA_group: str = "PA",
        SIH_group: str = "RD",
        PNI_group: str = "CPNI",
        local_dir: str = cache_dir,
    ) -> Union[tuple[str], str]:
        dbc_paths = self._get_dbc_paths(
            UFs=UFs,
            years=years,
            months=months,
            SINAN_disease=SINAN_disease,
            CNES_group=CNES_group,
            SIA_group=SIA_group,
            SIH_group=SIH_group,
            PNI_group=PNI_group,
        )
        data_dir = Path(local_dir)

        downloaded_parquets = []
        for path in dbc_paths:

            parquet_dir = data_dir / str(path).split("/")[-1].upper().replace(
                ".DBC", ".parquet"
            )

            if Path(parquet_dir).exists():
                downloaded_parquets.append(str(parquet_dir))
            else:
                local_filepath = self._extract_dbc(path, local_dir=local_dir)
                parquet_dir = self._dbfc_to_parquets(
                    local_filepath, local_dir=local_dir
                )
                downloaded_parquets.append(str(parquet_dir))
        return (
            downloaded_parquets[0] 
            if len(downloaded_parquets) == 1 
            else tuple(downloaded_parquets)
        )

    def _get_dbc_paths(
        self,
        UFs: Union[str, list] = None,
        years: Union[str, int, list] = None,
        months: Union[str, int, list] = None,
        SINAN_disease: str = None,
        CNES_group: str = None,
        SIA_group: str = "PA",
        SIH_group: str = "RD",
        PNI_group: str = "CPNI",
    ) -> list:
        parse_to_list = lambda ite: [ite] if not isinstance(ite, list) else ite
        UFs = parse_to_list(UFs)
        years = parse_to_list(years)
        months = parse_to_list(months)

        db = self._ftp_db.database
        list_files = self._ftp_db.list_all
        if db == "SINAN":
            all_dbcs = list_files(SINAN_disease=SINAN_disease)
            sinan_dis = FTP_SINAN(SINAN_disease)
        elif db == "CNES":
            all_dbcs = list_files(CNES_group=CNES_group)
        elif db == "SIA":
            all_dbcs = list_files(SIA_group=SIA_group)
        elif db == "SIH":
            all_dbcs = list_files(SIH_group=SIH_group)
        elif db == "PNI":
            all_dbcs = list_files(PNI_group=PNI_group)
        else:
            all_dbcs = list_files()

        def url_regex(
            month: str = None, year: str = None, UF: str = None
        ) -> re.Pattern:
            """
            Each url case is matched using regex patterns, mostly databases
            have the same file pattern, but some discrepancies can be found,
            for instance, lowercase UF and entire years and shortened years
            at the same time.
            """
            if db == "SINAN":
                if not year:
                    raise ValueError("Missing year(s)")
                file_pattern = re.compile(
                    f"{sinan_dis.code}BR{year}.dbc", re.I
                )
            elif db == "SIM" or db == "SINASC":
                if not year or not UF:
                    raise ValueError("Missing year(s) or UF(s)")
                file_pattern = re.compile(
                    rf"[DON]+R?{UF}\d?\d?{year}.dbc", re.I
                )
            elif db == "SIH":
                if not year or not month or not UF:
                    raise ValueError("Missing year(s), month(s) or UF(s)")
                file_pattern = re.compile(
                    rf"{SIH_group}{UF}{year}{month}.dbc", re.I
                )
            elif db == "SIA":
                if not year or not month or not UF:
                    raise ValueError("Missing year(s), month(s) or UF(s)")
                file_pattern = re.compile(
                    rf"{SIA_group}{UF}{year}{month}[abc]?.dbc", re.I
                )
            elif db == "PNI":
                if not year or not UF:
                    raise ValueError("Missing year(s) or UF(s)")
                file_pattern = re.compile(rf"{PNI_group}{UF}{year}.dbf", re.I)
            elif db == "CNES":
                if not year or not month or not UF:
                    raise ValueError("Missing year(s), month(s) or UF(s)")
                file_pattern = re.compile(
                    rf"{CNES_group}/{CNES_group}{UF}{year}{month}.dbc", re.I
                )
            elif db == "CIHA":
                if not year or not month or not UF:
                    raise ValueError("Missing year(s), month(s) or UF(s)")
                file_pattern = re.compile(rf"CIHA{UF}{year}{month}.dbc", re.I)
            return file_pattern
        files = list()
        for y, m, uf in product(
            years or [], months or [], UFs or []
        ):  # Allows None
            norm = lambda y: str(y)[-2:].zfill(2)
            regex = url_regex(year=norm(y), month=norm(m), UF=str(uf))
            filtered = list(filter(regex.search, all_dbcs))
            files.extend(filtered)
        return files

    def _extract_dbc(self, DBC_path: str, local_dir: str = cache_dir) -> str:
        """
        Enters in the FTP server and retrieve the DBC(F) path into
        local machine.
        """
        Path(local_dir).mkdir(exist_ok=True, parents=True)
        filename = DBC_path.split("/")[-1]
        filedir = DBC_path.replace(filename, "")
        filepath = Path(local_dir) / filename
        if Path(filepath).exists():
            return str(filepath)
        try:
            ftp = ftp = FTP("ftp.datasus.gov.br")
            ftp.login()
            ftp.cwd(filedir)
            ftp.retrbinary(
                f"RETR {filename}",
                open(f"{filepath}", "wb").write,
            )
            ftp.close()
            return str(filepath)
        except error_perm as e:
            logging.error(f"Not able to download {filename}")
            raise e

    def _dbfc_to_parquets(self, fpath: str, local_dir: str) -> str(PosixPath):
        """DBC/DBF files to parquets using Pandas & PyArrow"""
        db_path = Path(local_dir) / fpath
        dbfile = str(db_path.absolute()).split("/")[-1]
        if Path(dbfile).suffix in [".dbc", ".DBC"] and db_path.exists():
            outpath = f"{fpath[:-4]}.dbf"
            try:
                dbc2dbf(fpath, outpath)
                if Path(fpath).exists():
                    Path(fpath).unlink()
                fpath = outpath
            except Exception as e:
                logging.error(e)
                raise e

        parquet_dir = f"{fpath[:-4]}.parquet"
        if Path(parquet_dir).exists() and any(os.listdir(parquet_dir)):
            return parquet_dir
        Path(parquet_dir).mkdir(exist_ok=True, parents=True)

        def decode_column(value):
            # https://stackoverflow.com/questions/57371164/django-postgres-a-string-literal-cannot-contain-nul-0x00-characters
            if isinstance(value, bytes):
                return value.decode(encoding="iso-8859-1").replace("\x00", "")
            elif isinstance(value, str):
                return str(value).replace("\x00", "")
            else:
                return value

        for d in self._stream_DBF(DBF(fpath, encoding="iso-8859-1", raw=True)):
            try:
                df = pd.DataFrame(d)
                table = pa.Table.from_pandas(df.applymap(decode_column))
                pq.write_to_dataset(table, root_path=parquet_dir)

            except Exception as e:
                logging.error(e)

        if Path(fpath).exists():
            Path(fpath).unlink()

        return parquet_dir

    def _stream_DBF(self, dbf, chunk_size=30000):
        """Fetches records in chunks to preserve memory"""
        data = []
        i = 0
        for records in dbf:
            data.append(records)
            i += 1
            if i == chunk_size:
                yield data
                data = []
                i = 0
        else:
            yield data


class FTP_SINAN:
    name: str
    diseases: dict = {
        "Animais Peçonhentos": "ANIM",
        "Botulismo": "BOTU",
        "Cancer": "CANC",
        "Chagas": "CHAG",
        "Chikungunya": "CHIK",
        "Colera": "COLE",
        "Coqueluche": "COQU",
        "Contact Communicable Disease": "ACBI",
        "Acidentes de Trabalho": "ACGR",
        "Dengue": "DENG",
        "Difteria": "DIFT",
        "Esquistossomose": "ESQU",
        "Febre Amarela": "FAMA",
        "Febre Maculosa": "FMAC",
        "Febre Tifoide": "FTIF",
        "Hanseniase": "HANS",
        "Hantavirose": "HANT",
        "Hepatites Virais": "HEPA",
        "Intoxicação Exógena": "IEXO",
        "Leishmaniose Visceral": "LEIV",
        "Leptospirose": "LEPT",
        "Leishmaniose Tegumentar": "LTAN",
        "Malaria": "MALA",
        "Meningite": "MENI",
        "Peste": "PEST",
        "Poliomielite": "PFAN",
        "Raiva Humana": "RAIV",
        "Sífilis Adquirida": "SIFA",
        "Sífilis Congênita": "SIFC",
        "Sífilis em Gestante": "SIFG",
        "Tétano Acidental": "TETA",
        "Tétano Neonatal": "TETN",
        "Tuberculose": "TUBE",
        "Violência Domestica": "VIOL",
        "Zika": "ZIKA",
    }

    def __init__(self, name: str) -> None:
        self.name = self.__diseasecheck__(name)
        ftp = FTP_datasus()
        code = self.diseases[self.name]
        self.finals = ftp.nlst(f"{DB_PATHS['SINAN'][0]}/{code}BR*.dbc")
        self.prelims = ftp.nlst(f"{DB_PATHS['SINAN'][1]}/{code}BR*.dbc")

    def __diseasecheck__(self, name: str) -> str:
        return (
            name
            if name in self.diseases.keys()
            else ValueError(f"{name} not found.")
        )

    def __repr__(self) -> str:
        return f"SINAN Disease ({self.name})"

    def __str__(self) -> str:
        return self.name

    @property
    def code(self) -> str:
        return self.diseases[self.name]

    def get_years(self, stage: str = "all") -> list:
        """
        Returns the available years to download, if no stage
        is assigned, it will return years from both finals and
        preliminaries datasets.
        stage (str): 'finais' | 'prelim' | 'all'
        """

        def extract_years(paths):
            return [
                str(path).split("/")[-1].split(".dbc")[0][-2:]
                for path in paths
            ]

        prelim_years = extract_years(self.prelims)
        finais_years = extract_years(self.finals)

        if stage == "prelim":
            return sorted(prelim_years)
        elif stage == "finais":
            return sorted(finais_years)
        return sorted(prelim_years + finais_years)

    def get_ftp_paths(self, years: list) -> list:
        """
        Returns the FTP path available for years to download.
        years (list): a list with years to download, if year
                      is not available, it won't be included
                      in the result
        """
        prelim_paths = self.prelims
        finais_paths = self.finals
        all_paths = prelim_paths + finais_paths
        ds_paths = list()

        def mask(_year):
            return str(_year)[-2:].zfill(2)

        for year in years:
            [ds_paths.append(path) for path in all_paths if mask(year) in path]

        return ds_paths



"""
Download data from CIHA and CIH (Old)
Hospital and Ambulatorial information system
http://ciha.datasus.gov.br/CIHA/index.php?area=03

by fccoelho
license: GPL V3 or Later
"""
from typing import Union

from pysus.online_data import CACHEPATH, FTP_Downloader


def download(
    states: Union[str, list],
    years: Union[str, list, int],
    months: Union[str, list, int],
    data_dir: str = CACHEPATH,
) -> list:
    """
    Download CIHA records for state, year and month and returns dataframe
    :param months: 1 to 12, can be a list
    :param states: 2 letter state code,
    :param years: 4 digit integer
    """
    return FTP_Downloader("CIHA").download(
        UFs=states,
        years=years,
        months=months,
        local_dir=data_dir,
    )



import unittest
import pandas as pd
from pysus.online_data import IBGE


class SIDRA(unittest.TestCase):
    def test_get_aggregates(self):
        df = IBGE.list_agregados()
        self.assertIsInstance(df, pd.DataFrame)
        self.assertGreater(df.size, 0)

    def test_localidades_por_agregado(self):
        df = IBGE.localidades_por_agregado(475, nivel='N3')
        self.assertIsInstance(df, pd.DataFrame)
        self.assertGreater(df.size, 0)

    def test_lista_periodos(self):
        df = IBGE.lista_periodos(475)
        self.assertIsInstance(df, pd.DataFrame)
        self.assertGreater(df.size, 0)

    def test_get_sidra_table(self):
        df = IBGE.get_sidra_table(200, territorial_level=6, geocode=4220000, period='last', classification=2,
                                  categories='all')
        self.assertIsInstance(df, pd.DataFrame)
        self.assertGreater(df.size, 0)

    def test_metadata(self):
        md = IBGE.metadados(475)
        self.assertIsInstance(md, dict)
        self.assertGreater(len(md), 0)

    def test_FetchData(self):
        ds = IBGE.FetchData(475, periodos=1996, variavel=93, localidades='N3[all]',
                            classificacao='58[all]|2[4,5]|1[all]', view='flat')
        self.assertIsInstance(ds, IBGE.FetchData)
        self.assertGreater(len(ds.JSON), 0)


if __name__ == '__main__':
    unittest.main()



import unittest

import pandas as pd

from pysus.online_data.PNI import *
from pysus.online_data import parquets_to_dataframe


class PNITestCase(unittest.TestCase):
    def test_get_available_years(self):
        res = get_available_years("AC")
        self.assertIsInstance(res, list)
        self.assertIn('2000', res)

    def test_get_available_docs(self):
        res = available_docs()
        self.assertIsInstance(res, list)

    def test_download(self):
        files = download("RO", 2000)
        df = parquets_to_dataframe(files)
        self.assertIsInstance(df, pd.DataFrame)


if __name__ == "__main__":
    unittest.main()



import unittest

from pysus.online_data.ESUS import download


class MyTestCase(unittest.TestCase):
    @unittest.skip  # long test. Uncomment to run once.
    def test_download(self):
        df = download(uf="se")
        self.assertGreater(len(df), 0)


if __name__ == "__main__":
    unittest.main()



import unittest

from pysus.online_data.Infogripe import download, DATASETS


class InfoGripeTestCase(unittest.TestCase):
    def test_download(self):
        for ds in DATASETS.keys():
            df = download(ds)
            self.assertGreater(len(df), 0)  # add assertion here


if __name__ == '__main__':
    unittest.main()



"""
Download Mortality records from SIM Datasus
Created on 12/12/18
by fccoelho
license: GPL V3 or Later
"""
import os
from ftplib import FTP, error_perm
from typing import Union

import pandas as pd
from dbfread import DBF
from loguru import logger
from pysus.online_data import CACHEPATH, FTP_Downloader


def download(
    states: Union[str, list],
    years: Union[str, list, int],
    data_dir: str = CACHEPATH,
):
    """
    Downloads data directly from Datasus ftp server
    :param states: two-letter state identifier: MG == Minas Gerais
                   can be a list
    :param years: 4 digit integer, can be a list
    :return: a list of downloaded parquet paths
    """
    return FTP_Downloader("SIM").download(
        UFs=states, years=years, local_dir=data_dir
    )


def get_CID10_chapters_table(cache=True):
    """
    Fetch the CID10 chapters table
    :param cache: If set to True, stores data as parquets.
    :return: Pandas DataFrame
    """
    ftp = FTP("ftp.datasus.gov.br")
    ftp.login()
    logger.debug(
        f"Stablishing connection with ftp.datasus.gov.br.\n{ftp.welcome}"
    )
    ftp.cwd("/dissemin/publicos/SIM/CID10/TABELAS")
    logger.debug(
        "Changing FTP work dir to: /dissemin/publicos/SIM/CID10/TABELAS"
    )

    fname = "CIDCAP10.DBF"
    cachefile = os.path.join(
        CACHEPATH, "SIM_" + fname.split(".")[0] + "_.parquet"
    )

    if os.path.exists(cachefile):
        logger.info(f"Local parquet file found at {cachefile}")
        df = pd.read_parquet(cachefile)

        return df

    try:
        ftp.retrbinary("RETR {}".format(fname), open(fname, "wb").write)

    except error_perm:
        raise Exception("Could not download {}".format(fname))

    dbf = DBF(fname, encoding="iso-8859-1")
    df = pd.DataFrame(list(dbf))

    if cache:
        df.to_parquet(cachefile)
        logger.info(f"Data stored as parquet at {cachefile}")

    os.unlink(fname)
    logger.debug(f"{fname} removed")

    return df


def get_CID10_table(cache=True):
    """
    Fetch the CID10 table
    :param cache: If set to True, stores data as parquets.
    :return: Pandas DataFrame
    """
    ftp = FTP("ftp.datasus.gov.br")
    ftp.login()
    logger.debug(
        f"Stablishing connection with ftp.datasus.gov.br.\n{ftp.welcome}"
    )
    ftp.cwd("/dissemin/publicos/SIM/CID10/TABELAS")
    logger.debug(
        "Changing FTP work dir to: /dissemin/publicos/SIM/CID10/TABELAS"
    )

    fname = "CID10.DBF"
    cachefile = os.path.join(
        CACHEPATH, "SIM_" + fname.split(".")[0] + "_.parquet"
    )

    if os.path.exists(cachefile):
        logger.info(f"Local parquet file found at {cachefile}")
        df = pd.read_parquet(cachefile)

        return df

    try:
        ftp.retrbinary("RETR {}".format(fname), open(fname, "wb").write)

    except error_perm:
        raise Exception("Could not download {}".format(fname))

    dbf = DBF(fname, encoding="iso-8859-1")
    df = pd.DataFrame(list(dbf))

    if cache:
        df.to_parquet(cachefile)
        logger.info(f"Data stored as parquet at {cachefile}")

    os.unlink(fname)
    logger.debug(f"{fname} removed")

    return df


def get_CID9_table(cache=True):
    """
    Fetch the CID9 table
    :param cache: If set to True, stores data as parquets.
    :return: Pandas DataFrame
    """
    ftp = FTP("ftp.datasus.gov.br")
    ftp.login()
    logger.debug(
        f"Stablishing connection with ftp.datasus.gov.br.\n{ftp.welcome}"
    )
    ftp.cwd("/dissemin/publicos/SIM/CID9/TABELAS")
    logger.debug(
        "Changing FTP work dir to: /dissemin/publicos/SIM/CID9/TABELAS"
    )

    fname = "CID9.DBF"
    cachefile = os.path.join(
        CACHEPATH, "SIM_" + fname.split(".")[0] + "_.parquet"
    )

    if os.path.exists(cachefile):
        logger.info(f"Local parquet file found at {cachefile}")
        df = pd.read_parquet(cachefile)

        return df

    try:
        ftp.retrbinary("RETR {}".format(fname), open(fname, "wb").write)

    except error_perm:
        raise Exception("Could not download {}".format(fname))

    dbf = DBF(fname, encoding="iso-8859-1")
    df = pd.DataFrame(list(dbf))

    if cache:
        df.to_parquet(cachefile)
        logger.info(f"Data stored as parquet at {cachefile}")

    os.unlink(fname)
    logger.debug(f"{fname} removed")

    return df


def get_municipios(cache=True):
    """
    Get municipality metadata
    :param cache: If set to True, stores data as parquets.
    :return: Pandas DataFrame
    """
    ftp = FTP("ftp.datasus.gov.br")
    ftp.login()
    logger.debug(
        f"Stablishing connection with ftp.datasus.gov.br.\n{ftp.welcome}"
    )
    ftp.cwd("/dissemin/publicos/SIM/CID10/TABELAS")
    logger.debug(
        "Changing FTP work dir to: /dissemin/publicos/SIM/CID10/TABELAS"
    )

    fname = "CADMUN.DBF"
    cachefile = os.path.join(
        CACHEPATH, "SIM_" + fname.split(".")[0] + "_.parquet"
    )

    if os.path.exists(cachefile):
        logger.info(f"Local parquet file found at {cachefile}")
        df = pd.read_parquet(cachefile)

        return df

    try:
        ftp.retrbinary("RETR {}".format(fname), open(fname, "wb").write)

    except:
        raise Exception("Could not download {}".format(fname))

    dbf = DBF(fname, encoding="iso-8859-1")
    df = pd.DataFrame(list(dbf))

    if cache:
        df.to_parquet(cachefile)
        logger.info(f"Data stored as parquet at {cachefile}")

    os.unlink(fname)
    logger.debug(f"{fname} removed")

    return df


def get_ocupations(cache=True):
    """
    Fetch ocupations table
    :param cache: If set to True, stores data as parquets.
    :return: Pandas DataFrame
    """
    ftp = FTP("ftp.datasus.gov.br")
    ftp.login()
    logger.debug(
        f"Stablishing connection with ftp.datasus.gov.br.\n{ftp.welcome}"
    )
    ftp.cwd("/dissemin/publicos/SIM/CID10/TABELAS")
    logger.debug(
        "Changing FTP work dir to: /dissemin/publicos/SIM/CID10/TABELAS"
    )
    fname = "TABOCUP.DBF"
    cachefile = os.path.join(
        CACHEPATH, "SIM_" + fname.split(".")[0] + "_.parquet"
    )

    if os.path.exists(cachefile):
        logger.info(f"Local parquet file found at {cachefile}")
        df = pd.read_parquet(cachefile)

        return df

    try:
        ftp.retrbinary("RETR {}".format(fname), open(fname, "wb").write)

    except:
        raise Exception("Could not download {}".format(fname))

    dbf = DBF(fname, encoding="iso-8859-1")
    df = pd.DataFrame(list(dbf))

    if cache:
        df.to_parquet(cachefile)
        logger.info(f"Data stored as parquet at {cachefile}")

    os.unlink(fname)
    logger.debug(f"{fname} removed")

    return df



"""
Downloads SIA data from Datasus FTP server
Created on 21/09/18
by fccoelho
Modified on 22/11/22
by bcbernardo
license: GPL V3 or Later
"""
from pprint import pprint
from typing import Dict, Tuple, Union

from pysus.online_data import CACHEPATH, FTP_Downloader

group_dict: Dict[str, Tuple[str, int, int]] = {
    "PA": ("Produção Ambulatorial", 7, 1994),
    "BI": ("Boletim de Produção Ambulatorial individualizado", 1, 2008),
    "AD": ("APAC de Laudos Diversos", 1, 2008),
    "AM": ("APAC de Medicamentos", 1, 2008),
    "AN": ("APAC de Nefrologia", 1, 2008),
    "AQ": ("APAC de Quimioterapia", 1, 2008),
    "AR": ("APAC de Radioterapia", 1, 2008),
    "AB": ("APAC de Cirurgia Bariátrica", 1, 2008),
    "ACF": ("APAC de Confecção de Fístula", 1, 2008),
    "ATD": ("APAC de Tratamento Dialítico", 1, 2008),
    "AMP": ("APAC de Acompanhamento Multiprofissional", 1, 2008),
    "SAD": ("RAAS de Atenção Domiciliar", 1, 2008),
    "PS": ("RAAS Psicossocial", 1, 2008),
}


def show_datatypes():
    pprint(group_dict)


def download(
    states: Union[str, list],
    years: Union[str, list, int],
    months: Union[str, list, int],
    group: str = "PA",
    data_dir: str = CACHEPATH,
) -> list:
    """
    Download SIASUS records for state year and month and returns dataframe
    :param months: 1 to 12, can be a list
    :param states: 2 letter state code, can be a list
    :param years: 4 digit integer, can be a list
    :param data_dir: whether to cache files locally. default is True
    :param group: 2-3 letter document code, defaults to ['PA', 'BI'].
    Codes should be one of the following:
        PA - Produção Ambulatorial
        BI - Boletim de Produção Ambulatorial individualizado
        AD - APAC de Laudos Diversos
        AM - APAC de Medicamentos
        AN - APAC de Nefrologia
        AQ - APAC de Quimioterapia
        AR - APAC de Radioterapia
        AB - APAC de Cirurgia Bariátrica
        ACF - APAC de Confecção de Fístula
        ATD - APAC de Tratamento Dialítico
        AMP - APAC de Acompanhamento Multiprofissional
        SAD - RAAS de Atenção Domiciliar
        PS - RAAS Psicossocial
    :return: list of downloaded parquet paths
    """
    return FTP_Downloader("SIA").download(
        UFs=states,
        years=years,
        months=months,
        local_dir=data_dir,
        SIA_group=group,
    )



__author__ = "fccoelho"

import unittest

import pandas as pd

from pysus.online_data.SIH import download
from pysus.online_data import parquets_to_dataframe as to_df

unittest.skip("too slow to run om travis")


class SIHTestCase(unittest.TestCase):
    def test_download(self):
        df = to_df(download("to", 2009, 12))
        df2 = to_df(download("AC", 2013, 10))
        self.assertGreater(len(df), 0)
        self.assertGreater(len(df2), 0)
        self.assertIsInstance(df, pd.DataFrame)


if __name__ == "__main__":
    unittest.main()



from typing import Union

from pysus.online_data import CACHEPATH, FTP_Downloader

group_dict = {
    "LT": ["Leitos - A partir de Out/2005", 10, 2005],
    "ST": ["Estabelecimentos - A partir de Ago/2005", 8, 2005],
    "DC": ["Dados Complementares - A partir de Ago/2005", 8, 2005],
    "EQ": ["Equipamentos - A partir de Ago/2005", 8, 2005],
    "SR": ["Serviço Especializado - A partir de Ago/2005", 8, 2005],
    "HB": ["Habilitação - A partir de Mar/2007", 3, 2007],
    "PF": ["Profissional - A partir de Ago/2005", 8, 2005],
    "EP": ["Equipes - A partir de Abr/2007", 5, 2007],
    "IN": ["Incentivos - A partir de Nov/2007", 11, 2007],
    "RC": ["Regra Contratual - A partir de Mar/2007", 3, 2007],
    "EE": ["Estabelecimento de Ensino - A partir de Mar/2007", 3, 2007],
    "EF": ["Estabelecimento Filantrópico - A partir de Mar/2007", 3, 2007],
    "GM": ["Gestão e Metas - A partir de Jun/2007", 6, 2007],
}


def download(
    group: str,
    states: Union[str, list],
    years: Union[str, list, int],
    months: Union[str, list, int],
    data_dir: str = CACHEPATH,
) -> list:
    """
    Download CNES records for group, state, year and month and returns a
    list of local parquet files
    :param group:
        LT – Leitos - A partir de Out/2005
        ST – Estabelecimentos - A partir de Ago/2005
        DC - Dados Complementares - A partir de Ago/2005
        EQ – Equipamentos - A partir de Ago/2005
        SR - Serviço Especializado - A partir de Ago/2005
        HB – Habilitação - A partir de Mar/2007
        PF – Profissional - A partir de Ago/2005
        EP – Equipes - A partir de Abr/2007
        IN – Incentivos - A partir de Nov/2007
        RC - Regra Contratual - A partir de Mar/2007
        EE - Estabelecimento de Ensino - A partir de Mar/2007
        EF - Estabelecimento Filantrópico - A partir de Mar/2007
        GM - Gestão e Metas - A partir de Jun/2007
    :param months: 1 to 12, can be a list of years
    :param states: 2 letter state code, can be a list of UFs
    :param years: 4 digit integer, can be a list of years
    """
    return FTP_Downloader("CNES").download(
        CNES_group=group,
        UFs=states,
        years=years,
        months=months,
        local_dir=data_dir,
    )



# -*- coding:utf-8 -*-
u"""
Created on 23/09/2020
by gabrielmcf
license: GPL V3 or Later
"""

import unittest

from numpy.testing import (
    assert_array_almost_equal,
    assert_array_equal,
    assert_equal,
)

from pysus.online_data.SIM import download
from pysus.preprocessing import SIM, decoders
from pysus.online_data import parquets_to_dataframe as to_df


class TestDecoder(unittest.TestCase):
    def test_group_and_count(self):
        df = to_df(download("se", 2010))
        df = decoders.translate_variables_SIM(df)
        variables = ["CODMUNRES", "SEXO", "IDADE_ANOS"]
        counts = SIM.group_and_count(df, variables)
        self.assertGreater(counts.COUNTS.sum(), 0)

    def test_redistribute_missing(self):
        df = to_df(download("se", 2010))
        df = decoders.translate_variables_SIM(df)
        variables = ["CODMUNRES", "SEXO", "IDADE_ANOS"]
        counts = SIM.group_and_count(df, variables)
        sum_original = counts["COUNTS"].sum()
        counts = SIM.redistribute_missing(counts, variables)
        sum_redistributed = counts["COUNTS"].sum()

        self.assertEqual(sum_original, sum_redistributed)



    def test_redistribute_missing_partial(self):
        df = to_df(download("se", 2010))
        df = decoders.translate_variables_SIM(
            df, age_classes=True, classify_cid10_chapters=True
        )
        group_variables = ["CODMUNRES", "SEXO", "IDADE_ANOS", "CID10_CHAPTER"]
        counts = SIM.group_and_count(df, group_variables)
        counts["COUNTS_ORIGINAL"] = counts["COUNTS"]
        sum_original = counts["COUNTS"].sum()
        counts = SIM.redistribute_missing(counts, group_variables[:3])
        sum_redistributed = counts["COUNTS"].sum()

        assert_equal(sum_original, round(sum_redistributed))
        # sample = counts[counts['COUNTS'] != 0]['COUNTS'].sample(20,random_state=0).tolist()
        # assert_array_almost_equal(sample, [1.0026605509150972, 3.0076529330337682, 10.0, 3.0, 1.0, 7.030611240693058, 2.0, 1.0, 1.0003988761766138, 1.0, 5.0, 1.0, 2.0, 1.0, 1.0011890475332716, 1.0007766913402458, 3.0, 3.0, 1.0, 1.0], decimal=5)



"""
Downloads SIH data from Datasus FTP server
Created on 21/09/18
by fccoelho
license: GPL V3 or Later
"""
from typing import Union

from pysus.online_data import CACHEPATH, FTP_Downloader


def download(
    states: Union[str, list],
    years: Union[str, list, int],
    months: Union[str, list, int],
    data_dir: str = CACHEPATH,
) -> list:
    """
    Download SIH records for state year and month and returns dataframe
    :param months: 1 to 12, can be a list
    :param states: 2 letter state code, can be alist
    :param years: 4 digit integer, can be a list
    :param data_dir: Directory where parquets will be downloaded.
    :return: a list of parquet paths
    """
    return FTP_Downloader("SIH").download(
        UFs=states,
        years=years,
        months=months,
        SIH_group="RD",
        local_dir=data_dir,
    )



# -*- coding:utf-8 -*-
u"""
This module contains a set of functions to process data on SIM
Created on 16/08/2020
by gabrielmcf
license: GPL V3 or Later
"""

__docformat__ = "restructuredtext en"
from decimal import Decimal
from itertools import product

import numpy as np
import pandas as pd


def logical_and_from_dict(dataframe, dictionary):
    if dictionary == {}:
        return np.array([True] * len(dataframe), dtype=bool)
    return np.logical_and.reduce([dataframe[k] == v for k, v in dictionary.items()])


def relax_filter(dictionary, fields):
    for field in reversed(fields):
        if field in dictionary:
            del dictionary[field]
            break
    return dictionary


def group_and_count(
    dataframe, group_columns, count_column="COUNTS", decimal_counts=False
):
    """
    Agrupa e conta as variáveis passadas como parâmetro no dataframe. Cria uma nova
    coluna de contagem, com o tipo Decimal para possibilitar redistribuição pro rata posterior e maior precisão.
    :param dataframe: dataframe pandas
    :param group_columns: lista de string contendo o nome das colunas a serem agrupadas no dataframe.
    :param count_columns: nome da coluna de counts a ser criada.
    :return:
    """
    counts = dataframe.groupby(group_columns).size().reset_index(name=count_column)

    if decimal_counts:
        counts[count_column] = counts[count_column].apply(lambda x: Decimal(x))
    else:
        counts[count_column] = counts[count_column].astype("float64")

    return counts


def redistribute_missing(
    counts, filter_columns, count_column="COUNTS", nan_string="nan"
):
    """
    Realiza redistribuição pro rata das contagens do SIM com algum dado faltante.
    O dataframe deve conter uma coluna float64 chamada CONTAGEM e as demais colunas devem ser
    do tipo category, tendo os dados faltantes em uma categoria definida pelo parâmetro nan_string.
    :param counts: dataframe pandas contendo coluna com soma chamada CONTAGEM
    :param filter_columns: variáveis a serem consideradas para filtro de redistribuição pro rata
    :param count_columns: nome da coluna de counts.
    :param nan_string: string usada na categoria de dado faltante
    :return:
    """

    # Removendo categorias faltantes vazias
    for var in filter_columns:
        condition_dict = {var: nan_string, count_column: 0.0}
        counts = counts[~logical_and_from_dict(counts, condition_dict)]

    ### Dataframes de dados faltantes

    variables_dict = [{x: nan_string} for x in filter_columns]

    variables_condition = [logical_and_from_dict(counts, x) for x in variables_dict]

    # Primeiro item da tupla é != nan, segundo é o == nan
    variables_tuples = [(np.logical_not(x), x) for x in variables_condition]
    variables_product = list(product(*variables_tuples))

    # Remove regra de todos != nan
    del variables_product[0]

    # Lista todos os dados faltantes por grupos de colunas faltantes
    list_missing_data = [counts[np.logical_and.reduce(x)] for x in variables_product]
    # Remove as colunas de dado faltante dos dataframes
    list_missing_data = [
        x.drop(columns=x.columns[x.isin([nan_string]).any()].tolist())
        for x in list_missing_data
    ]
    # Remove os conjuntos vazios
    list_missing_data = list(filter(lambda x: not x.empty, list_missing_data))

    # Remove dados faltantes
    counts = counts[~np.logical_or.reduce(variables_product[-1])]

    # Executa para cada conjunto de dados faltantes
    for missing_count in list_missing_data:
        counts = redistribute_rows_pro_rata(counts, filter_columns, missing_count)

    return counts


def redistribute_cid_chapter(
    counts,
    filter_columns,
    chapter=18,
    chapter_column="CID10_CHAPTER",
    count_columns="COUNTS",
):
    """
    Realiza redistribuição pro rata das contagens do SIM de um capítulo do CID10 passado.
    Por padrão o capítulo XVIII, de causas mal definidas, é redistribuído.
    :param counts: dataframe pandas contendo coluna com contagem
    :param filter_columns: variáveis a serem consideradas para filtro na redistribuição pro rata
    :param chapter: capítulo do CID10 a ser redistribuído
    :param chapter_column: nome da coluna de capítulo
    :param count_columns: nome da coluna de counts
    :return:
    """
    df_chapter = counts[
        (counts[chapter_column] == chapter) & (counts[count_columns] > 0)
    ]
    counts = counts[counts[chapter_column] != chapter]

    return redistribute_rows_pro_rata(counts, filter_columns, df_chapter)


def redistribute_rows_pro_rata(
    counts, filter_columns, redistribute_list, count_columns="COUNTS"
):
    """
    Redistribui as contagens do dataframe conforme as colunas de filtro passadas.
    :param counts: dataframe pandas contendo coluna de contagem
    :param filter_columns: variáveis a serem consideradas para filtro na redistribuição pro rata
    :param redistribute_list: dataframe contendo as linhas que serão redistribuídas
    :param count_columns: nome da coluna de counts
    :return:
    """
    # Evita alerta na atribuição de múltiplos itens com máscara (.loc)
    pd.set_option("mode.chained_assignment", None)

    not_filter_columns = list(set(counts.columns.to_list()) - set(filter_columns))

    for row in redistribute_list.itertuples(index=False):
        row_dict = dict(row._asdict())
        [row_dict.pop(key) for key in not_filter_columns]
        condition = logical_and_from_dict(counts, row_dict)
        sum_data = counts[condition][count_columns].sum()
        # Caso não haja proporção conhecida relaxa o filtro
        while sum_data == 0.0 and len(row_dict) > 0:
            row_dict = relax_filter(row_dict, filter_columns)
            condition = logical_and_from_dict(counts, row_dict)
            sum_data = counts[condition][count_columns].sum()
        counts.loc[condition, count_columns] = counts[condition][count_columns].apply(
            lambda x: pro_rata_model(x, getattr(row, count_columns), sum_data)
        )

    # Volta alerta para warning
    pd.set_option("mode.chained_assignment", "warn")
    return counts


def pro_rata_model(current_value, redistribution_amount, group_sum):
    return redistribution_amount * current_value / group_sum + current_value
